@@ -1,15 +1,15 @@
 #include "mold.h"
-#include "../common/cmdline.h"
 
 #include <random>
 #include <regex>
 #include <sstream>
 #include <sys/stat.h>
 #include <sys/types.h>
+#include <tbb/global_control.h>
 #include <unordered_set>
 
 #ifdef _WIN32
-# define _isatty isatty
+# define isatty _isatty
 # define STDERR_FILENO (_fileno(stderr))
 #else
 # include <unistd.h>
@@ -34,7 +34,7 @@ Options:
     --no-dynamic-linker
   -L DIR, --library-path DIR  Add DIR to library search path
   -M, --print-map             Write map file to stdout
-  -N, --omagic                Do not page align data, do not make text readonly
+  -N, --omagic                Do not page align data; do not make text readonly
     --no-omagic
   -O NUMBER                   Ignored
   -S, --strip-debug           Strip .debug_* sections
@@ -56,22 +56,25 @@ Options:
                               Trace references to SYMBOL
   --Bdynamic, --dy            Link against shared libraries (default)
   --Bstatic, --dn, --static   Do not link against shared libraries
-  --Bsymbolic                 Bind global symbols locally
-  --Bsymbolic-functions       Bind global functions locally
-  --Bno-symbolic              Cancel --Bsymbolic and --Bsymbolic-functions
+  --Bsymbolic                 Bind all symbols locally
+  --Bsymbolic-functions       Bind function symbols locally
+  --Bsymbolic-non-weak        Bind all but weak symbols locally
+  --Bsymbolic-non-weak-functions
+                              Bind all but weak function symbols locally
+  --Bno-symbolic              Cancel --Bsymbolic options
   --Map FILE                  Write map file to a given file
   --Tbss=ADDR                 Set address to .bss
-  --Tdata                     Set address to .data
-  --Ttext                     Set address to .text
+  --Tdata=ADDR                Set address to .data
+  --Ttext=ADDR                Set address to .text
   --allow-multiple-definition Allow multiple definitions
-  --apply-dynamic-relocs      Apply link-time values for dynamic relocations (defualt)
+  --apply-dynamic-relocs      Apply link-time values for dynamic relocations (default)
     --no-apply-dynamic-relocs
   --as-needed                 Only set DT_NEEDED if used
     --no-as-needed
   --build-id [none,md5,sha1,sha256,uuid,HEXSTRING]
                               Generate build ID
     --no-build-id
-  --chroot DIR                Set a given path to root directory
+  --chroot DIR                Set a given path to the root directory
   --color-diagnostics=[auto,always,never]
                               Use colors in diagnostics
   --color-diagnostics         Alias for --color-diagnostics=always
@@ -86,12 +89,12 @@ Options:
     --disable-new-dtags       Emit DT_RPATH for --rpath
   --execute-only              Make executable segments unreadable
   --dp                        Ignored
-  --dynamic-list              Read a list of dynamic symbols (implies -Bsymbolic)
+  --dynamic-list=FILE         Read a list of dynamic symbols (implies -Bsymbolic)
   --eh-frame-hdr              Create .eh_frame_hdr section
     --no-eh-frame-hdr
-  --exclude-libs LIB,LIB,..   Mark all symbols in given libraries hidden
+  --exclude-libs LIB,LIB,..   Mark all symbols in given libraries as hidden
   --export-dynamic-symbol     Put symbols matching glob in the dynamic symbol table
-  --export-dynamic-symbol-list
+  --export-dynamic-symbol-list=FILE
                               Read a list of dynamic symbols
   --fatal-warnings            Treat warnings as errors
     --no-fatal-warnings       Do not treat warnings as errors (default)
@@ -113,54 +116,55 @@ Options:
     --no-nmagic
   --no-undefined              Report undefined symbols (even with --shared)
   --noinhibit-exec            Create an output file even if errors occur
-  --oformat=binary            Omit ELF, section and program headers
+  --oformat=binary            Omit ELF, section, and program headers
   --pack-dyn-relocs=[relr,none]
                               Pack dynamic relocations
   --package-metadata=STRING   Set a given string to .note.package
   --perf                      Print performance statistics
-  --pie, --pic-executable     Create a position independent executable
+  --pie, --pic-executable     Create a position-independent executable
     --no-pie, --no-pic-executable
-  --pop-state                 Restore state of flags governing input file handling
+  --pop-state                 Restore the state of flags governing input file handling
   --print-gc-sections         Print removed unreferenced sections
     --no-print-gc-sections
   --print-icf-sections        Print folded identical sections
     --no-print-icf-sections
-  --push-state                Save state of flags governing input file handling
+  --push-state                Save the state of flags governing input file handling
   --quick-exit                Use quick_exit to exit (default)
     --no-quick-exit
   --relax                     Optimize instructions (default)
     --no-relax
-  --repro                     Embed input files to .repro section
+  --repro                     Embed input files in .repro section
   --require-defined SYMBOL    Require SYMBOL be defined in the final output
   --retain-symbols-file FILE  Keep only symbols listed in FILE
-  --reverse-sections          Reverses input sections in the output file
+  --reverse-sections          Reverse input sections in the output file
   --rosegment                 Put read-only non-executable sections in their own segment (default)
     --no-rosegment            Put read-only non-executable sections in an executable segment
-  --rpath DIR                 Add DIR to runtime search path
+  --rpath DIR                 Add DIR to the runtime search path
   --rpath-link DIR            Ignored
   --run COMMAND ARG...        Run COMMAND with mold as /usr/bin/ld
-  --section-start=SECTION=ADDR Set address to section
-  --shared, --Bshareable      Create a share library
+  --section-start=SECTION=ADDR Set address for section
+  --shared, --Bshareable      Create a shared library
   --shuffle-sections[=SEED]   Randomize the output by shuffling input sections
   --sort-common               Ignored
   --sort-section              Ignored
-  --spare-dynamic-tags NUMBER Reserve given number of tags in .dynamic section
+  --spare-dynamic-tags NUMBER Reserve the given number of tags in the .dynamic section
   --spare-program-headers NUMBER
-                              Reserve given number of slots in the program header
+                              Reserve the given number of slots in the program header
   --start-lib                 Give following object files in-archive-file semantics
     --end-lib                 End the effect of --start-lib
   --stats                     Print input statistics
-  --sysroot DIR               Set target system root directory
+  --sysroot DIR               Set the target system root directory
   --thread-count COUNT, --threads=COUNT
                               Use COUNT number of threads
   --threads                   Use multiple threads (default)
     --no-threads
-  --trace                     Print name of each input file
-  --undefined-version         Do not report version scripts that refer undefined symbols
-    --no-undefined-version    Report version scripts that refer undefined symbols (default)
+  --trace                     Print the name of each input file
+  --undefined-glob PATTERN    Force to resolve all symbols that match a given pattern
+  --undefined-version         Do not report version scripts that refer to undefined symbols
+    --no-undefined-version    Report version scripts that refer to undefined symbols (default)
   --unique PATTERN            Don't merge input sections that match a given pattern
   --unresolved-symbols [report-all,ignore-all,ignore-in-object-files,ignore-in-shared-libs]
-                              How to handle unresolved symbols
+                              Handle unresolved symbols
   --version-script FILE       Read version script
   --warn-common               Warn about common symbols
     --no-warn-common
@@ -172,21 +176,21 @@ Options:
                               Report unresolved symbols as errors (default)
   --whole-archive             Include all objects from static archives
     --no-whole-archive
-  --wrap SYMBOL               Use wrapper function for a given symbol
+  --wrap SYMBOL               Use a wrapper function for a given symbol
   -z defs                     Report undefined symbols (even with --shared)
     -z nodefs
   -z common-page-size=VALUE   Ignored
-  -z execstack                Require executable stack
+  -z execstack                Require an executable stack
     -z noexecstack
-  -z execstack-if-needed      Make the stack area execuable if an input file explicitly requests it
+  -z execstack-if-needed      Make the stack area executable if an input file explicitly requests it
   -z initfirst                Mark DSO to be initialized first at runtime
-  -z interpose                Mark object to interpose all DSOs but executable
+  -z interpose                Mark object to interpose all DSOs but the executable
   -z keep-text-section-prefix Keep .text.{hot,unknown,unlikely,startup,exit} as separate sections in the final binary
     -z nokeep-text-section-prefix
   -z lazy                     Enable lazy function resolution (default)
   -z max-page-size=VALUE      Use VALUE as the memory page size
   -z nocopyreloc              Do not create copy relocations
-  -z nodefaultlib             Make the dynamic loader to ignore default search paths
+  -z nodefaultlib             Make the dynamic loader ignore default search paths
   -z nodelete                 Mark DSO non-deletable at runtime
   -z nodlopen                 Mark DSO not available to dlopen
   -z nodump                   Mark DSO not available to dldump
@@ -196,11 +200,13 @@ Options:
     -z nopack-relative-relocs
   -z sectionheader            Do not omit section header (default)
     -z nosectionheader        Omit section header
+  -z start_stop_visibility=[hidden,protected]
+                              Specify symbol visibility for "__start_SECNAME" and "__stop_SECNAME" symbols
   -z separate-loadable-segments
-                              Separate all loadable segments to different pages
-    -z separate-code          Separate code and data into different pages
+                              Separate all loadable segments onto different pages
+    -z separate-code          Separate code and data onto different pages
     -z noseparate-code        Allow overlap in pages
-  -z stack-size=VALUE         Set size of stack segment
+  -z stack-size=VALUE         Set the size of the stack segment
   -z relro                    Make some sections read-only after relocation (default)
     -z norelro
   -z text                     Report error if DT_TEXTREL is set
@@ -210,6 +216,108 @@ Options:
 mold: supported targets: elf32-i386 elf64-x86-64 elf32-littlearm elf64-littleaarch64 elf32-littleriscv elf32-bigriscv elf64-littleriscv elf64-bigriscv elf32-powerpc elf64-powerpc elf64-powerpc elf64-powerpcle elf64-s390 elf64-sparc elf32-m68k elf32-sh-linux elf64-alpha elf64-loongarch elf32-loongarch
 mold: supported emulations: elf_i386 elf_x86_64 armelf_linux_eabi aarch64linux aarch64elf elf32lriscv elf32briscv elf64lriscv elf64briscv elf32ppc elf32ppclinux elf64ppc elf64lppc elf64_s390 elf64_sparc m68kelf shlelf_linux elf64alpha elf64loongarch elf32loongarch)";
 
+template <typename E>
+static std::vector<std::string_view>
+read_response_file(Context<E> &ctx, std::string_view path, i64 depth) {
+  if (depth > 10)
+    Fatal(ctx) << path << ": response file nesting too deep";
+
+  std::vector<std::string_view> vec;
+  MappedFile *mf = must_open_file(ctx, std::string(path));
+  std::string_view data((char *)mf->data, mf->size);
+
+  while (!data.empty()) {
+    if (isspace(data[0])) {
+      data = data.substr(1);
+      continue;
+    }
+
+    auto read_quoted = [&]() {
+      char quote = data[0];
+      data = data.substr(1);
+
+      std::string buf;
+      while (!data.empty() && data[0] != quote) {
+        if (data[0] == '\\' && data.size() >= 1) {
+          buf.append(1, data[1]);
+          data = data.substr(2);
+        } else {
+          buf.append(1, data[0]);
+          data = data.substr(1);
+        }
+      }
+      if (data.empty())
+        Fatal(ctx) << path << ": premature end of input";
+      data = data.substr(1);
+      return save_string(ctx, buf);
+    };
+
+    auto read_unquoted = [&] {
+      std::string buf;
+      while (!data.empty()) {
+        if (data[0] == '\\' && data.size() >= 1) {
+          buf.append(1, data[1]);
+          data = data.substr(2);
+          continue;
+        }
+
+        if (!isspace(data[0])) {
+          buf.append(1, data[0]);
+          data = data.substr(1);
+          continue;
+        }
+        break;
+      }
+      return save_string(ctx, buf);
+    };
+
+    std::string_view tok;
+    if (data[0] == '\'' || data[0] == '\"')
+      tok = read_quoted();
+    else
+      tok = read_unquoted();
+
+    if (tok.starts_with('@'))
+      append(vec, read_response_file(ctx, tok.substr(1), depth + 1));
+    else
+      vec.push_back(tok);
+  }
+  return vec;
+}
+
+// Replace "@path/to/some/text/file" with its file contents.
+template <typename E>
+std::vector<std::string_view>
+expand_response_files(Context<E> &ctx, char **argv) {
+  std::vector<std::string_view> vec;
+  for (i64 i = 0; argv[i]; i++) {
+    if (argv[i][0] == '@')
+      append(vec, read_response_file(ctx, argv[i] + 1, 1));
+    else
+      vec.push_back(argv[i]);
+  }
+  return vec;
+}
+
+static i64 get_default_thread_count() {
+  // mold doesn't scale well above 32 threads.
+  int n = tbb::global_control::active_value(
+    tbb::global_control::max_allowed_parallelism);
+  return std::min(n, 32);
+}
+
+static inline std::string_view string_trim(std::string_view str) {
+  size_t pos = str.find_first_not_of(" \t");
+  if (pos == str.npos)
+    return "";
+  str = str.substr(pos);
+
+  pos = str.find_last_not_of(" \t");
+  if (pos == str.npos)
+    return str;
+  return str.substr(0, pos + 1);
+}
+
 static std::vector<std::string> add_dashes(std::string name) {
   // Single-letter option
   if (name.size() == 1)
@@ -297,8 +405,7 @@ split_by_comma_or_colon(std::string_view str) {
 
 template <typename E>
 static void read_retain_symbols_file(Context<E> &ctx, std::string_view path) {
-  MappedFile<Context<E>> *mf =
-    MappedFile<Context<E>>::must_open(ctx, std::string(path));
+  MappedFile *mf = must_open_file(ctx, std::string(path));
   std::string_view data((char *)mf->data, mf->size);
 
   ctx.arg.retain_symbols_file.reset(new std::unordered_set<std::string_view>);
@@ -437,7 +544,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
     ctx.arg.apply_dynamic_relocs = false;
 
   auto read_arg = [&](std::string name) {
-    for (std::string opt : add_dashes(name)) {
+    for (const std::string &opt : add_dashes(name)) {
       if (args[0] == opt) {
         if (args.size() == 1)
           Fatal(ctx) << "option -" << name << ": argument missing";
@@ -457,7 +564,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
   };
 
   auto read_eq = [&](std::string name) {
-    for (std::string opt : add_dashes(name)) {
+    for (const std::string &opt : add_dashes(name)) {
       if (args[0].starts_with(opt + "=")) {
         arg = args[0].substr(opt.size() + 1);
         args = args.subspan(1);
@@ -468,7 +575,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
   };
 
   auto read_flag = [&](std::string name) {
-    for (std::string opt : add_dashes(name)) {
+    for (const std::string &opt : add_dashes(name)) {
       if (args[0] == opt) {
         args = args.subspan(1);
         return true;
@@ -519,13 +626,13 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
     } else if (read_flag("no-dynamic-linker")) {
       ctx.arg.dynamic_linker = "";
     } else if (read_flag("v")) {
-      SyncOut(ctx) << mold_version;
+      SyncOut(ctx) << get_mold_version();
       version_shown = true;
     } else if (read_flag("version")) {
-      SyncOut(ctx) << mold_version;
+      SyncOut(ctx) << get_mold_version();
       exit(0);
     } else if (read_flag("V")) {
-      SyncOut(ctx) << mold_version
+      SyncOut(ctx) << get_mold_version()
                    << "\n  Supported emulations:\n   elf_x86_64\n   elf_i386\n"
                    << "   aarch64linux\n   armelf_linux_eabi\n   elf64lriscv\n"
                    << "   elf64briscv\n   elf32lriscv\n   elf32briscv\n"
@@ -580,12 +687,15 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
     } else if (read_flag("no-export-dynamic")) {
       ctx.arg.export_dynamic = false;
     } else if (read_flag("Bsymbolic")) {
-      ctx.arg.Bsymbolic = true;
+      ctx.arg.Bsymbolic = BSYMBOLIC_ALL;
     } else if (read_flag("Bsymbolic-functions")) {
-      ctx.arg.Bsymbolic_functions = true;
+      ctx.arg.Bsymbolic = BSYMBOLIC_FUNCTIONS;
+    } else if (read_flag("Bsymbolic-non-weak")) {
+      ctx.arg.Bsymbolic = BSYMBOLIC_NON_WEAK;
+    } else if (read_flag("Bsymbolic-non-weak-functions")) {
+      ctx.arg.Bsymbolic = BSYMBOLIC_NON_WEAK_FUNCTIONS;
     } else if (read_flag("Bno-symbolic")) {
-      ctx.arg.Bsymbolic = false;
-      ctx.arg.Bsymbolic_functions = false;
+      ctx.arg.Bsymbolic = BSYMBOLIC_NONE;
     } else if (read_arg("exclude-libs")) {
       append(ctx.arg.exclude_libs, split_by_comma_or_colon(arg));
     } else if (read_flag("q") || read_flag("emit-relocs")) {
@@ -670,6 +780,9 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
         Fatal(ctx) << "unknown --unresolved-symbols argument: " << arg;
     } else if (read_arg("undefined") || read_arg("u")) {
       ctx.arg.undefined.push_back(get_symbol(ctx, arg));
+    } else if (read_arg("undefined-glob")) {
+      if (!ctx.arg.undefined_glob.add(arg, 0))
+        Fatal(ctx) << "--undefined-glob: invalid pattern: " << arg;
     } else if (read_arg("require-defined")) {
       ctx.arg.require_defined.push_back(get_symbol(ctx, arg));
     } else if (read_arg("init")) {
@@ -832,9 +945,10 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
       ctx.page_size = parse_number(ctx, "-z max-page-size", arg);
       if (!has_single_bit(ctx.page_size))
         Fatal(ctx) << "-z max-page-size " << arg << ": value must be a power of 2";
-    } else if (read_z_arg("start-stop-visibility")) {
-      if (arg != "hidden")
-        Fatal(ctx) << "-z start-stop-visibility: unsupported visibility: " << arg;
+    } else if (read_z_flag("start-stop-visibility=protected")) {
+      ctx.arg.z_start_stop_visibility_protected = true;
+    } else if (read_z_flag("start-stop-visibility=hidden")) {
+      ctx.arg.z_start_stop_visibility_protected = false;
     } else if (read_z_flag("noexecstack")) {
       ctx.arg.z_execstack = false;
     } else if (read_z_flag("relro")) {
@@ -1109,6 +1223,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
     } else if (read_flag("warn-constructors")) {
     } else if (read_flag("warn-execstack")) {
     } else if (read_flag("no-warn-execstack")) {
+    } else if (read_flag("long-plt")) {
     } else if (read_flag("secure-plt")) {
     } else if (read_arg("rpath-link")) {
     } else if (read_z_flag("combreloc")) {
@@ -1126,7 +1241,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
       // cannot be done right here.
       remaining.push_back("--version-script=" + std::string(arg));
     } else if (read_arg("dynamic-list")) {
-      ctx.arg.Bsymbolic = true;
+      ctx.arg.Bsymbolic = BSYMBOLIC_ALL;
       append(ctx.dynamic_list_patterns, parse_dynamic_list(ctx, arg));
     } else if (read_arg("export-dynamic-symbol")) {
       ctx.dynamic_list_patterns.push_back({arg, "<command line>"});
@@ -1296,6 +1411,7 @@ std::vector<std::string> parse_nonpositional_args(Context<E> &ctx) {
 
 using E = MOLD_TARGET;
 
+template std::vector<std::string_view> expand_response_files(Context<E> &, char **);
 template std::vector<std::string> parse_nonpositional_args(Context<E> &ctx);
 
 } // namespace mold::elf
